/*
 * 2D Polygon Landscape
 *
 * Copyright (c) 2011  Tobias Zwick
 */

#include <cstdlib>
#define _USE_MATH_DEFINES
#include <cmath>
#include <algorithm>
#include <cassert>

#include "Polygon.h"
#include "ringvec.h"
#include "PolygonClipper.h"

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* ++++++++++++++++++++++++++++++ Constructors +++++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

Polygon::Polygon(const Polygon &poly) : points(poly.points), _position(poly._position), _size(poly._size)
{
	_convex = poly._convex;
	_area = poly._area;
	_perimeter = poly._perimeter;
	calculateStartIterator();
}

Polygon::Polygon(Vertex arr[], int length) : points(arr,arr+length)
{
	_convex = checkConvexity();
	calculateSize();
	calculatePerimeter();
	calculateBoundingBox();
	calculateStartIterator();
}

Polygon::Polygon(const std::list<Vertex> &newpoints)
{
	points.reserve(newpoints.size());
	for(std::list<Vertex>::const_iterator it = newpoints.begin(); it != newpoints.end(); ++it)
	{
		points.push_back(Vertex(*it));
	}
	_convex = checkConvexity();
	calculateSize();
	calculatePerimeter();
	calculateBoundingBox();
	calculateStartIterator();
}

Polygon::Polygon(const std::vector<Vertex> &newpoints) : points(newpoints)
{
	_convex = checkConvexity();
	calculateSize();
	calculatePerimeter();
	calculateBoundingBox();
	calculateStartIterator();
}



/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* ++++++++++++++++++++ Checks Executed On Construction ++++++++++++++++++++ */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void Polygon::calculateStartIterator()
{
	_startit = points.begin();

	// the left-most and lowest vertex is the start vertex
	for(std::vector<Vertex>::const_iterator it = points.begin(); it != points.end(); ++it)
	{
		// choose left-most
		if(it->x() < _startit->x())
			_startit = it;
		else if(it->x() == _startit->x())
		{
			// if there are several: lower-most
			if(it->y() < _startit->y())
				_startit = it;
			// if there are still several: same for next next
			else if(it->y() == _startit->y())
			{
				// (must be weakly simple polygon)
				// same with next
				std::vector<Vertex>::const_iterator it_n = increaseIt(it,points);
				std::vector<Vertex>::const_iterator best_n = increaseIt(_startit,points);
				
				// left-most
				if(it_n->x() < best_n->x())
					_startit = it;
				else if(it_n->x() == best_n->x())
				{
					// lowest
					if(it_n->y() < best_n->y())
						_startit = it;
				}
			}
		}
	}
}

bool Polygon::checkConvexity() const
{
	// all simple polygon whith less than 4 points are convex
	if(points.size() < 4) return true;

	for(std::vector<Vertex>::const_iterator it = points.begin(); it != points.end(); ++it)
	{
		Line previous(*decreaseIt(it,points,2),*decreaseIt(it,points,1));

		// if this point is right of the previous line, we can terminate here
		// because the polygon is not convex
		if(previous.getDirection(*it) > 0) return false;
	}
	return true;
}

void Polygon::calculateBoundingBox()
{
	// if there are no points, there is no bounding box
	if(points.size() == 0)
	{
		_size = Vertex();
		_position = Vertex();
		return;
	}

	std::vector<Vertex>::const_iterator it = points.begin();

	// get the leftmost, rightmost, uppermost and bottommost vertex
	Vertex upperright = *it;
	Vertex bottomleft = *it;

	while(it != points.end())
	{
		if(it->x() < bottomleft.x()) bottomleft.x(it->x());
		if(it->x() > upperright.x()) upperright.x(it->x());
		if(it->y() < bottomleft.y()) bottomleft.y(it->y());
		if(it->y() > upperright.y()) upperright.y(it->y());

		++it;
	}
	_size = upperright-bottomleft;
	// position is the bottom-left corner
	_position = bottomleft;
}

void Polygon::calculateSize()
{
	// this is just using the Gaussian trapezoid formula
	int sum = 0;

	for(std::vector<Vertex>::const_iterator it = points.begin(); it != points.end(); ++it)
	{
		Vertex A = *decreaseIt(it,points);
		Vertex B = *it;

		sum += (A.y() + B.y()) * (A.x() - B.x());
	}
	// for ccw defined polygons, the above returns a positive value,
	// for cw negative. We don't want to loose that information here
	//sum = abs(sum);

	// actually, the area is almost an integer, were it not for that
	// / 2.0 that could result in something.5 ... :-)
	_area = ((double) sum / 2.0);
}

void Polygon::calculatePerimeter()
{
	double sum = 0;
	// the sum of all lines in the polygon...
	for(std::vector<Vertex>::const_iterator it = points.begin(); it != points.end(); ++it)
	{
		Line line(Vertex(*decreaseIt(it,points)),Vertex(*it),Line::Segment);
		sum += line.getLength();
	}
	_perimeter = sum;
}


/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* ++++++++++++++++++++++++++++ Simple Operators +++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

Polygon& Polygon::operator=(const Polygon &poly)
{
	points = poly.points;
	_position = poly._position;
	_size = poly._size;
	_convex = poly._convex;
	_area = poly._area;
	_perimeter = poly._perimeter;
	calculateStartIterator();
	return *this;
}

Polygon Polygon::operator-() const
{
	std::vector<Vertex> newpoints;
	newpoints.reserve(points.size());
	for(std::vector<Vertex>::const_reverse_iterator it = points.rbegin(); it != points.rend(); ++it)
	{
		newpoints.push_back(*it);
	}
	return Polygon(newpoints);
}

bool Polygon::operator==(const Polygon& poly) const
{
	// other polygon doesn't have the same number of vertices
	if(points.size() != poly.points.size()) return false;

	std::vector<Vertex>::const_iterator it = _startit;
	std::vector<Vertex>::const_iterator jt = poly._startit;

	do 
	{
		// point is not the same -> false
		if(*it != *jt) return false;
		it = increaseIt(it,points);
		jt = increaseIt(jt,poly.points);
	} while(it != _startit);

	return true;
}

bool Polygon::operator!=(const Polygon& poly) const
{
	return !(*this == poly);
}




/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++++++++++ Checks ++++++++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

const std::vector<Vertex>& Polygon::vertices() const
{
	return points;
}

bool Polygon::convex() const
{
	return _convex;
}

double Polygon::size() const
{
	return _area;
}

double Polygon::perimeter() const
{
	return _perimeter;
}

bool Polygon::valid() const
{
	// less than 3 vertices?
	if(points.size() < 3) return false;

	// zero or smaller size
	if(size() <= 0) return false;

	// line segments of length 0?
	std::vector<Vertex>::const_iterator it = points.begin();
	++it;
	while(it != points.end())
	{
		std::vector<Vertex>::const_iterator before = it;
		--before;
		if(*before == *it)
			return false;
		++it;
	}

	// self-intersections
	for(std::vector<Vertex>::const_iterator it1 = points.begin(); it1 != points.end();++it1)
	{
		Line edge1(*decreaseIt(it1,points),*it1,Line::Segment);
		std::vector<Vertex>::const_iterator it2 = it1;
		for(++it2; it2 != points.end();++it2)
		{
			Line edge2(*decreaseIt(it2,points),*it2,Line::Segment);
			Intersection intersect;
			if(edge1.intersects(edge2,intersect))
			{
				if(intersect.isLineIntersection()) return false;
				// next/prev line: its normal to intersect with last point
				if(increaseIt(it1,points) == it2)
					if(intersect.thisParam == 1.0) continue;
				if(decreaseIt(it1,points) == it2)
					if(intersect.thisParam == 0.0) continue;
				return false;
			}
		}
	}

	return true;
}



/* +++++++++++++++++++++++++++ Point Containment +++++++++++++++++++++++++++ */

bool Polygon::contains(const Vertex& point) const
{
	if(!insideBoundingBox(point)) return false;

	/*
	Check if one point is in the polygon. For non-convex polygons, we need
	to check the following:

	Each line segment goes from A to B. We start with the last vertex 
	as A and the first vertex as B and then iterate.

	We count the intersections of a ray going from the given point 
	with all line segments of the polygon. If the number is odd, it's
	inside, otherwise it's outside

	---- 

	The method for convex polygons is simpler and little faster, especially
	for points that are not in the polygon after all (because unlike the other
	algorithm, it doesn't always need to search all lines but stops on the one
	that doesn't conform.
	Method: The point has to lie on the right side of each line segment of the
	polygon.
	BUT: The Polygon must be defined in a counter clockwise order.

	*/

	int intersections = 0;
	// it is 2x faster to compute with a ray that is on one of the axes
	Line ray(point,point+Vertex(0,100),Line::Ray);
	int dirBefore = 0;

	for(std::vector<Vertex>::const_iterator it = points.begin(); it != points.end(); ++it)
	{
		Vertex A = *decreaseIt(it,points);
		Vertex B = *it;
		Line edge(A,B,Line::Segment);

		// for convex polygons
		if(_convex)
		{
			if(edge.getDirection(point) > 0) return false;
		}
		// for concave polygons
		else
		{
			// the ray intersects a polygon edge
			Intersection intersect;
			if(ray.intersects(edge, intersect))
			{
				// optimization: if the starting point of the ray (= the point) is 
				// actually on an edge of the polygon, the point is inside.
				if(intersect.thisParam == 0) return true;

				if(!intersect.isLineIntersection())
				{
					// the first point does not belong to the edge (otherwise we would
					// find every point twice here) that's why it is skipped
					if(intersect.otherParam == 0) continue;

					// special case: the ray hit a vertex (vertex B)
					if(intersect.otherParam == 1)
					{
						Vertex C = *increaseIt(it,points);
						// check position of A and C
						int dirA = ray.getDirection(A);
						int dirC = ray.getDirection(C);
						// no zeros (it means at next pos there is a line intersection)
						if(dirC == 0) continue;
						// on different sides? success!!
						if((dirA > 0) ^ (dirC > 0))
							++intersections;
					}
					// otherwise, the ray crosses the boundary of the polygon
					else ++intersections;
				}
				// very special case: the ray hit an edge
				else
				{
					Vertex C = *increaseIt(it,points);
					Vertex Z = *decreaseIt(it,points,2);
					int dirC = ray.getDirection(C);
					// memorize last direction change
					int dirZ = ray.getDirection(Z);
					if(dirZ != 0) dirBefore = dirZ;
					// uber special case: no zeros (it means at next pos there is even another line intersection)
					if(dirC == 0) continue;
					// on different sides? success!!
					if((dirBefore > 0) ^ (dirC > 0))
						++intersections;
				}
			}
		}
	}
	// either it is convex and all edges are on the right side of the point
	// or the crossings count is odd -> true
	return _convex || (intersections % 2 == 1);
}


bool Polygon::insideBoundingBox(const Vertex& point) const
{
	// is inside enclosing rectangle?
	if(point.x() >= _position.x())
		if(point.y() >= _position.y())
			if(point.x() <= _position.x()+_size.x())
				if(point.y() <= _position.y()+_size.y())
					return true;
	return false;
}

bool Polygon::insideBoundingBox(const Polygon &poly) const
{
	Line mine = getRectangleLine();
	Line other = poly.getRectangleLine();
	if(other.B().x() >= mine.A().x())
		if(mine.B().x() >= other.A().x())
			if(other.B().y() >= mine.A().y())
				if(mine.B().y() >= other.A().y())
					return true;
	return false;
}

Line Polygon::getRectangleLine() const
{
	return Line(_position,_position + _size);
}



/* ++++++++++++++++++++++++ Intersection With Line +++++++++++++++++++++++++ */

bool Polygon::intersects(const Line& line, std::list<Intersection> &intersections, bool backfaceCulling) const
{
	// check if line is in bounding box:
	// it's only worth to check if its in the rectangle first if there are much more lines
	// the minimum number of 4 is chosen arbitrary by me (checking if the line intersects
	// with the rect is about worth 2-3 intersections )
	if(points.size() >= 4)
		if(!intersectsBoundingBox(line))
			return false;

	// extra data for convex polygons
	unsigned int intersectCount = 0;
	bool foundLineIntersection = false;

	// for each line in the polygon...
	for(std::vector<Vertex>::const_iterator it = points.begin(); it != points.end(); ++it)
	{
		Vertex A = *decreaseIt(it,points);
		Vertex B = *it;
		Line polyLine(A,B,Line::Segment);

		// backface culling
		if(backfaceCulling)
		{
			if(polyLine.getNormal() * line.getVector() > 0)
			{
				continue;
			}
		}

		// intersection found!
		Intersection intersection;
		if(polyLine.intersects(line, intersection))
		{

			// intersection is a point
			if(!intersection.isLineIntersection())
			{
				// first point of the edge doesn't count
				if(intersection.thisParam == 0.0) continue;

				intersections.push_back(intersection);

				// lines are not counted as intersections for this break condition
				// because in convex polygons there can still be an unlimited number of
				// line intersections.
				intersectCount++;
			}
			// a line
			else
			{
				intersections.push_back(intersection);
				foundLineIntersection = true;
			}
		}
		// if the polygon is convex, in the current iteration there is no
		// intersection and before one has been found, we can stop here
		else {
			if(_convex)
			{
				if(foundLineIntersection)
					break;
				if(backfaceCulling ? intersectCount > 0 : intersectCount > 1)
					break;
			}
		} // end if-else intersection...

	} // end for each point

	return intersectCount != 0 || foundLineIntersection;
}

bool Polygon::intersectsBoundingBox(const Line& line) const
{

	// if one of the points is inside the rect it intersects the rect
	if(insideBoundingBox(line.A())) return true;
	if(insideBoundingBox(line.B())) return true;

	// otherwise calculate the intersections...
	return !calculateIntersectionsWithBoundingBox(line,true).empty();
}

std::list<Intersection> Polygon::calculateIntersectionsWithBoundingBox(const Line& line, bool stop_if_found_one) const
{
	std::vector<Vertex> vertices;
	// bottomleft
	vertices.push_back(_position);
	// bottomright
	vertices.push_back(Vertex(_position.x()+_size.x(), _position.y()));
	// upperright
	vertices.push_back(_position + _size);
	// upperleft
	vertices.push_back(Vertex(_position.x(), _position.y()+_size.y()));

	std::list<Intersection> intersections;

	for(std::vector<Vertex>::const_iterator it = vertices.begin(); it != vertices.end(); ++it)
	{
		Line edge(*decreaseIt(it,vertices),*it,Line::Segment);
		Intersection intersect;
		if(line.intersects(edge,intersect))
		{
			intersections.push_back(Intersection(intersect));
			if(stop_if_found_one)	return intersections;
		}
	}
	return intersections;
}


/* +++++++++++++++++++++++++++++++++ Move ++++++++++++++++++++++++++++++++ */

void Polygon::move(const Vertex& by)
{
	for(std::vector<Vertex>::iterator it = points.begin(); it != points.end(); ++it)
	{
		*it += by;
	}
	_position += by;
}



/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* ++++++++++++++++++++++++++ Clipping operations ++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

std::list<Polygon*> Polygon::operator-(const Polygon &poly) const
	{ return clipOperation(poly, Difference); }
std::list<Polygon*> Polygon::operator|(const Polygon &poly) const
	{ return clipOperation(poly,Union); }
std::list<Polygon*> Polygon::operator&(const Polygon &poly) const
	{ return clipOperation(poly,Intersect); }
std::list<Polygon*> Polygon::operator^(const Polygon &poly) const
	{ return clipOperation(poly,Exclusion); }

std::list<Polygon*> Polygon::clipOperation(const Polygon &poly, Clipping_operation op) const
{
	// wrapper for clip(...) which just adds copies of the original polygons to
	// the result if nothing has been changed
	std::list<Polygon*> result;
	Clip_result cr = clip(poly,op, result);
	if(cr & SubjectPolygon)
		result.push_back(new Polygon(*this));
	if(cr & ClipPolygon)
		result.push_back(new Polygon(poly));

	return result;
}

Clip_result Polygon::clip(const Polygon &poly, Clipping_operation op, std::list<Polygon*> &result) const
{
	PolygonClipper clipper(*this,poly);
	Clip_result clip_result = clipper.clip(op,result);
	return clip_result;
}



/* +++++++++++++++++++++++++++ Split polygons ++++++++++++++++++++++++++++++ */

std::list<Polygon*> Polygon::operator/(const Line& line) const
{
	std::list<Polygon*> result;
	Clip_result cr = split(line,result);
	if(cr == SubjectPolygon)
		result.push_back(new Polygon(*this));

	return result;

}

Clip_result Polygon::split(const Line& oline, std::list<Polygon*> &result) const
{
	assert(oline.getLength() > 0);

	Line line(oline);
	line.setType(Line::Segment);
	Line up(Vertex(0,0),Vertex(0,100));
	double angle = up.getAngle(line);

	std::vector<Vertex> first;
	std::vector<Vertex> second;
	first.reserve(6);
	second.reserve(6);

	// line goes more vertically than horizontally
	if(abs(angle) < M_PI/4.0 || abs(angle) > 3.0*M_PI/4.0)
	{
		// flip that it always goes up
		if(abs(angle) > M_PI/2.0) line = -line;
		Line orig(line);

		Vertex vector = line.getVector();

		// expand line till bigger than bounding box
		while(line.A().y() > _position.y())
			line = Line(line.A() - vector, line.B(), line.getType());
		while(line.B().y() < _position.y()+_size.y())
			line = Line(line.A(), line.B() + vector, line.getType());

		// construct polygons
		int right = std::max(std::max( line.A().x() , line.B().x() ), _position.x()+_size.x() ) + 1;
		int left =  std::min(std::min( line.A().x() , line.B().x() ), _position.x() ) - 1;

		first.push_back(line.B());
		// if the line has been expanded, still include the original points
		// this is necessary to guarantee that the original split line is not
		// altered by rounding imprecision
		if(orig.B() != line.B())
			first.push_back(orig.B());
		if(orig.A() != line.A())
			first.push_back(orig.A());
		first.push_back(line.A());
		first.push_back(Vertex(right, line.A().y()));
		first.push_back(Vertex(right, line.B().y()));

		second.push_back(line.A());
		if(orig.A() != line.A())
			second.push_back(orig.A());
		if(orig.B() != line.B())
			second.push_back(orig.B());
		second.push_back(line.B());
		second.push_back(Vertex(left, line.B().y()));
		second.push_back(Vertex(left, line.A().y()));
	}
	// line goes more horizontally than vertically
	else
	{
		// flip so that it always goes right
		if(angle < 0.0) line = -line;
		Line orig(line);

		Vertex vector = line.getVector();

		// expand line till bigger than bounding box
		while(line.A().x() > _position.x())
			line = Line(line.A() - vector, line.B(), line.getType());
		while(line.B().x() < _position.x()+_size.x())
			line = Line(line.A(), line.B() + vector, line.getType());

		// construct polygons
		int top =     std::max(std::max( line.A().y() , line.B().y() ), _position.y()+_size.y() ) + 1;
		int bottom =  std::min(std::min( line.A().y() , line.B().y() ), _position.y() ) - 1;

		first.push_back(line.B());
		// if the line has been expanded, still include the original points
		// this is necessary to guarantee that the original split line is not
		// altered by rounding imprecision
		if(orig.B() != line.B())
			first.push_back(orig.B());
		if(orig.A() != line.A())
			first.push_back(orig.A());
		first.push_back(line.A());
		first.push_back(Vertex(line.A().x(), bottom));
		first.push_back(Vertex(line.B().x(), bottom));

		second.push_back(line.A());
		if(orig.A() != line.A())
			second.push_back(orig.A());
		if(orig.B() != line.B())
			second.push_back(orig.B());
		second.push_back(line.B());
		second.push_back(Vertex(line.B().x(), top));
		second.push_back(Vertex(line.A().x(), top));
	}


	if(clip(Polygon(first),Intersect,result) != Default) return SubjectPolygon;
	// (if on one side of the split line, there is something else than the
	// whole (or no) polygon, on the other side there will be something, too)
	// so an additional shortcut check is not needed here.
	clip(Polygon(second),Intersect,result);

	return Default;
}




