/*
 * 2D Polygon Landscape
 *
 * Copyright (c) 2011  Tobias Zwick
 */

#ifndef _POLYGON_H_
#define _POLYGON_H_

#include <vector>
#include <list>
#include "Vertex.h"
#include "Line.h"

// the different clipping operations
enum Clipping_operation { Difference, Union, Exclusion, Intersect };
// the different results from a clipping operation (useable as bitmask)
enum Clip_result { Default=0, SubjectPolygon=1, ClipPolygon=2, BothPolygons=3, Nothing=4 };

// used by the polygon2d class
struct VertexComp {
	bool operator() (const Vertex& lhs, const Vertex& rhs) const
	{ return lhs.x()<rhs.x() || (lhs.y()<rhs.y() && lhs.x()==rhs.x()); }
};

/* Polygon
   Represents a polygon in 2D space

*/
class PolygonClipper;
class Polygon
{
public:
	Polygon(const Polygon &poly);
	Polygon(Vertex arr[], int length);
	Polygon(const std::list<Vertex> &newpoints);
	Polygon(const std::vector<Vertex> &newpoints);

	const std::vector<Vertex>& vertices() const;

	/* Returns true if polygon is convex. */
	bool convex() const;

	/* Returns true if the given point is inside the polygon. Runs in O(n). */
	bool contains(const Vertex& point) const;

	/* Returns true if the given line intersects with the polygon. The
	   intersection data is returned in intersections. If backfaceCulling
		 is true, only the edges that face to the line's direction vector
		 are checked. Runs in O(n). */
	bool intersects(const Line& line, std::list<Intersection> &intersections, bool backfaceCulling = false) const;

	/* Returns the size of the polygon. */
	double size() const;

	/* Returns the perimeter of the polygon */
	double perimeter() const;

	/* Move polygon by the given vector. */
	void move(const Vertex& by);

	/* Get the difference of this polygon and the clip polygon (A and not(B)) */
	std::list<Polygon*> operator-(const Polygon &poly) const;
	/* Get the union of this polygon and the clip polygon (A or B) */
	std::list<Polygon*> operator|(const Polygon &poly) const;
	/* Get the intersection of this polygon and the clip polygon (A and B)*/
	std::list<Polygon*> operator&(const Polygon &poly) const;
	/* Get the exclusion of this polygon and the clip polygon (A xor B) */
	std::list<Polygon*> operator^(const Polygon &poly) const;

	/* Divide the polygon into two (or more) polygons along the given line */
	std::list<Polygon*> operator/(const Line& line) const;

	/* assign value of other polygon */
	Polygon& operator=(const Polygon &poly);
	/* compare two polygons */
	bool operator==(const Polygon &poly) const;
	bool operator!=(const Polygon &poly) const;
	/* get the negative version of this polygon (counter-clockwise <-> clockwise defined) */
	Polygon operator-() const;

	/* Perform a clipping operation on this polygon with the given polygon.
	   Possible are Difference, Union, Intersect and Exclusion. The result
		 contains only newly created polygons; if the polygons remain unmodified,
		 the method returns an appropiate clip_result other than Default. */
	Clip_result clip(const Polygon &poly, Clipping_operation op, std::list<Polygon*> &result) const;
	/* Perform a splittiing operation on this polygon with the given line.
	   The result contains only newly created polygons; if the polygons remain
		 unmodified, the method returns an appropiate clip_result other than Default.
		 Apart from this, this is the same operation as divide(line). */
	Clip_result split(const Line& line, std::list<Polygon*> &result) const;

	bool valid() const;

protected:
	// polygon points
	std::vector<Vertex> points;

	Line getRectangleLine() const;

private:
	/* ++++++++++++++++++++++++++++++ Variables ++++++++++++++++++++++++++++++ */

	// enclosing rectangle
	Vertex _position;
	Vertex _size;

	// convexity
	bool _convex;

	// area
	double _area;

	// start iterator - same for all "permutations" of the same polygon
	std::vector<Vertex>::const_iterator _startit;

	// perimeter
	double _perimeter;

	/* +++++++++++++++++++++++++++ Private Methods +++++++++++++++++++++++++++ */

	std::list<Polygon*> clipOperation(const Polygon &poly, Clipping_operation op) const;

	/* other */

	bool insideBoundingBox(const Vertex &point) const;
	bool insideBoundingBox(const Polygon &poly) const;
	bool intersectsBoundingBox(const Line& line) const;
	std::list<Intersection> calculateIntersectionsWithBoundingBox(const Line& line, bool stop_if_found_one) const;

	bool checkConvexity() const;

	void calculateSize();
	void calculatePerimeter();
	void calculateStartIterator();
	void calculateBoundingBox();

	friend class PolygonClipper;
};

#endif

